import math
import time
import numpy as np
import matplotlib.path as mplPath

def enumerate(graph, crossings, shapes_included):    
    def vector(p1, p2):

        """
        Creates a vector from points p1 and p2
        """

        return (p2[0] - p1[0], p2[1] - p1[1])

    def dot(p1, p2):

        """
        Finds the dot product of points p1 and p2
        """

        return p1[0]*p2[0] + p1[1]*p2[1]

    def mag(p):

        """
        Finds the magnitude of points p1 and p2
        """

        return math.sqrt(p[0]**2 + p[1]**2)

    def validAngles(cycle):

        """
        Checks that every angle in cycle is the same value
        """

        # No angle is formed with less than 3 points
        if len(cycle) <= 3:
            return True

        p1,p2,p3,p4 = cycle[-4],cycle[-3],cycle[-2],cycle[-1]
        vectorBA, vectorBC = vector(p2,p1), vector(p2,p3)
        t1 = math.acos(round(dot(vectorBA,vectorBC)/(mag(vectorBA)*mag(vectorBC)),4))
        vectorBA, vectorBC = vector(p3,p2), vector(p3,p4)
        t2 = math.acos(round(dot(vectorBA,vectorBC)/(mag(vectorBA)*mag(vectorBC)),4))

        return t1 == t2

    threes = set()
    def addThrees(cycle):

        """
        Adds every set of 3 cyclically-consecutive points and
        it's reverse in cycle
        """

        for i in range(len(cycle[:-2])):    
            threes.add((cycle[i], cycle[i+1], cycle[i+2]))
            threes.add((cycle[i+2], cycle[i+1], cycle[i]))

        threes.add((cycle[-2], cycle[-1], cycle[0]))
        threes.add((cycle[0], cycle[-1], cycle[-2]))

        threes.add((cycle[-1], cycle[0], cycle[1]))
        threes.add((cycle[1], cycle[0], cycle[-1]))

    def identical(cycle):

        """
        Checks if cycle is cyclically identical to
        any other cycles already added
        """

        if len(cycle) < 3:
            return False
        
        return (cycle[-3], cycle[-2], cycle[-1]) in threes

    def identicalLast(cycle):

        """
        Checks if the last points in cycle are in threes
        Same as identical(cycle) except only looks at sets
        of three that wrap around to the first elements
        """

        return (cycle[-2], cycle[-1], cycle[0]) in threes or (cycle[-1], cycle[0], cycle[1]) in threes

    #https://stackoverflow.com/questions/40833612/find-all-cycles-in-a-graph-implementation
    #Modified for efficiency
    def dfs(graph, start, end):

        """
        Depth first search through the directed graph.
        Optimized using properties of k-uniform tilings
        and the specific tiling being drawn
        """
            
        fringe = [(start, [])]
        while fringe:

            state, path = fringe.pop()

            #Ignores potential cycle if angle properties
            #aren't follow or if it is identical to a cycle
            #that has already been added
            if not validAngles(path) or identical(path):

                continue

            #If the cycle found is the correct length
            #and is not identical to another cycle, 
            #it gets added to cycles
            if state == end and len(path) in shapes_included and not identicalLast(path):

                #Adds new sets of three
                addThrees(path)

                yield path
                continue

            for next_state in graph[state]:

                #If the current path plus 1 is longer than 
                #the shape with the most edges in the tiling
                #the path isn't considered any more
                if next_state in path or len(path) + 1 > max(shapes_included):

                    continue

                fringe.append((next_state, path+[next_state]))
        
    start = time.process_time()

    #Enumerates every face
    cycles = [path for node in graph for path in dfs(graph, node, node)]

    #Removes hexagons and dodecagons that are made of other shapes
    for cycle in list(cycles):

        #Ignores triangles, squares and octagons
        #which cannot be made of other shapes
        if len(cycle) in [3,4,8]:

            continue

        #Hexagons can only be made of triangles that all contain
        #the centroid of the hexagon. If the centroid of the
        #hexagon is in crossings then the cycle is removed
        elif len(cycle) == 6:

            centroidX = round((cycle[0][0] + cycle[1][0] + cycle[2][0] + cycle[3][0] + cycle[4][0] + cycle[5][0])/6,3)
            centroidY = round((cycle[0][1] + cycle[1][1] + cycle[2][1] + cycle[3][1] + cycle[4][1] + cycle[5][1])/6,3)
            if (centroidX, centroidY) in crossings.keys():
                cycles.remove(cycle)
        
        else:
            #Dodecagons may have triangles in the center 
            #that contain the centroid of the dodecagon. 
            #If the centroid is in crossings then the 
            #cycle is removed
            centroidX = round((cycle[0][0] + cycle[1][0] + cycle[2][0] + cycle[3][0] + cycle[4][0] + cycle[5][0] + cycle[6][0] + cycle[7][0] + cycle[8][0] + cycle[9][0] + cycle[10][0] + cycle[11][0])/12,3)
            centroidY = round((cycle[0][1] + cycle[1][1] + cycle[2][1] + cycle[3][1] + cycle[4][1] + cycle[5][1] + cycle[6][1] + cycle[7][1] + cycle[8][1] + cycle[9][1] + cycle[10][1] + cycle[11][1])/12,3)
            if (centroidX, centroidY) in crossings.keys():

                cycles.remove(cycle)

            else:

                #If the dodecagon encloses any point in crossings that
                #aren't in the cycle then it is made of other shapes
                #and is removed from cycles
                poly_path = mplPath.Path(np.array(cycle))
                for pt in crossings:

                    if pt in cycle:

                        continue

                    if poly_path.contains_point(pt):

                        cycles.remove(cycle)
                        break


    print("Faces found: ", len(cycles))
    print(time.process_time() - start)
    return cycles